#!/usr/bin/env python


# Custom module imports
import config
import data
import scenes


from cocos.actions import *
from cocos.director import director
from pyglet import window

import cocos
import pyglet
import random
import sys


def load_levels():
    return [cocos.tiles.load(data.get_path("level_test.tmx"))["Tile Layer 1"]]


class MainMenuBackground(cocos.layer.Layer):

    is_event_handler = True

    def __init__(self):
        super(MainMenuBackground, self).__init__()

        background = cocos.sprite.Sprite(pyglet.image.load(data.get_path("europe.png")))
        background.position = int(config.SCREEN_WIDTH / 2), int(config.SCREEN_HEIGHT / 2)

        label = cocos.text.Label(
            "Napoleon Bonaparte is invading your country. You must not let him in!",
            font_name = "Calibri",
            font_size = 16,
            anchor_x = "center", anchor_y = "center"
        )
        label.position = int(config.SCREEN_WIDTH / 2), 130

        label2 = cocos.text.Label(
            "Use the arrow keys to navigate the menu and select a level and its difficulty",
            font_name = "Calibri",
            font_size = 12,
            anchor_x = "center", anchor_y = "center"
        )
        label2.position = int(config.SCREEN_WIDTH / 2), 110

        label3 = cocos.text.Label(
            "In game, at each turn, you will first place your reinforcement troops, attack, and then watch the enemy fight back.  You may also press ESCAPE at any time to go back to this menu.",
            font_name = "Calibri",
            font_size = 12,
            anchor_x = "center", anchor_y = "center"
        )
        label3.position = int(config.SCREEN_WIDTH / 2), 90

        label4 = cocos.text.Label(
            "To place reinforcements, just click on one of your own tiles you would like to add troops to.",
            font_name = "Calibri",
            font_size = 12,
            anchor_x = "center", anchor_y = "center"
        )
        label4.position = int(config.SCREEN_WIDTH / 2), 70

        label5 = cocos.text.Label(
            "To attack, just click on your tile you want to attack with and then click on a neighboring enemy cell to attack.  After attacking, press ENTER.",
            font_name = "Calibri",
            font_size = 12,
            anchor_x = "center", anchor_y = "center"
        )
        label5.position = int(config.SCREEN_WIDTH / 2), 50

        self.add(background)
        self.add(label)
        self.add(label2)
        self.add(label3)
        self.add(label4)
        self.add(label5)


class MainMenu(cocos.menu.Menu):
    def __init__(self, scene_manager, title = ""):
        super(MainMenu, self).__init__(title = title)
        self.scene_manager = scene_manager

        self.level_id = 1
        self.level_difficulty = 1

        l = []
        l.append(cocos.menu.MenuItem("Start Game", self.on_start_game))
        # l.append(cocos.menu.MultipleMenuItem("Level: ", self.on_level_difficulty, ["1"], default_item = self.level_id - 1))
        l.append(cocos.menu.MultipleMenuItem("Level Difficulty: ", self.on_level_difficulty, ["1", "2", "3", "4"], default_item = self.level_difficulty - 1))
        l.append(cocos.menu.MenuItem("Quit Game", self.on_quit))
        self.create_menu(l, cocos.menu.shake(), cocos.menu.shake_back())

    def on_start_game(self):
        self.scene_manager.do_level(self.level_id, self.level_difficulty)

    def on_quit(self):
        # GOODBYE!!!  :)
        sys.exit()

    def nothing(self):
        pass

    def on_level(self, index):
        self.level_id = index + 1

    def on_level_difficulty(self, index):
        self.level_difficulty = index + 1


class MainMenuScene(cocos.scene.Scene):
    def __init__(self, scene_manager, title = ""):
        super(MainMenuScene, self).__init__()

        self.scene_manager = scene_manager

        self.main_menu = MainMenu(self.scene_manager, title)

        self.add(MainMenuBackground())
        self.add(self.main_menu)


class TroopLayer(cocos.layer.Layer):
    def __init__(self):
        super(TroopLayer, self).__init__()

    def start(self):
        self.generate_labels()

    def get_level_layer(self, level_layer):
        self.level_layer = level_layer

    def generate_labels(self):
        '''label_x = cocos.text.Label(
            "1",
            font_name = "Times New Roman",
            font_size = 16,
            anchor_x = "center", anchor_y = "center"
        )'''
        self.troop_labels = [[cocos.text.Label("", position = self.level_layer.hex_map.cells[i][j].get_center(), font_name = "Calibri", font_size = 16) for j in range(0, len(self.level_layer.hex_map.cells[0]))] for i in range(0, len(self.level_layer.hex_map.cells))]

        # set the labels to their corresponding cell positions
        for i in range(0, len(self.troop_labels)):
            for j in range(0, len(self.troop_labels[i])):
                label = self.troop_labels[i][j]
                cell_center = self.level_layer.hex_map.cells[i][j].get_center()
                label.position = (cell_center[0] + 8, cell_center[1] + 14)
                self.add(label)

        self.available_troops = cocos.text.Label("", position = (8, 8), font_name = "Calibri", font_size = 32)
        self.add(self.available_troops)

    def update_labels(self):
        for i in range(0, len(self.troop_labels)):
            for j in range(0, len(self.troop_labels[i])):
                self.troop_labels[i][j].element.text = str(self.level_layer.hex_map.cells[i][j].properties["troops"])

        if self.level_layer.reinforcement_troops > 0:
            self.available_troops.element.text = "Available reinforcements: " + str(self.level_layer.reinforcement_troops)

        else:
            self.available_troops.element.text = ""


class LevelLayer(cocos.layer.ScrollingManager):

    is_event_handler = True

    def __init__(self, scene_manager, hex_map, difficulty):
        super(LevelLayer, self).__init__()

        self.scene_manager = scene_manager
        self.hex_map = hex_map
        # difficulty range from 1 to 4 (30%, 45%, 60%, or 75% of map taken by enemy)
        self.difficulty = difficulty

        self.add(self.hex_map)

        # variables!

        self.enemy_color = (100, 100, 100)
        self.focused_color = (155, 155, 155)

        self.difficulty_range = [.25, .5, .7, .8]
        self.difficulty_modifier = self.difficulty_range[self.difficulty - 1]
        self.reinforcement_troops_modifier = .5
        self.troops_placed = False
        self.random_range = 2
        # modifier of the number of troops in attack_A
        self.attack_strength = .5
        self.focused_cell = None

        # 0 -> self place reinforcement troops
        # 1 -> self attack
        # 2 -> enemy turn
        self.mode = 0

        self.reinforcement_troops = 0

        self.attack_A = None
        self.attack_B = None

    def start(self):
        # let the games begin!
        self.generate_territories()
        self.do_enemy_reinforcements()
        self.reinforcement_troops = self.get_reinforcement_troops("self")
        self.schedule(self.update)

    def get_troop_layer(self, troop_layer):
        self.troop_layer = troop_layer

    def on_key_press (self, key, modifiers):
        '''
            'key' is a constant indicating which key was pressed.
            'modifiers' is a bitwise or of several constants indicating which
                modifiers are active at the time of the press (ctrl, shift, capslock, etc.)
        '''

        # ESCAPE - Back to main menu
        if key == window.key.ESCAPE:
            self.scene_manager.do_main_menu()
            return True

        # ENTER - End turn
        elif key == window.key.ENTER:
            if self.mode == 1:
                self.mode = 2

    def on_mouse_press (self, x, y, buttons, modifiers):
        '''
            'buttons' is a bitwise or of pyglet.window.mouse constants LEFT, MIDDLE, RIGHT
            'modifiers' is a bitwise or of pyglet.window.key modifier constants
               (values like 'SHIFT', 'OPTION', 'ALT')
        '''

        target_hex = self.hex_map.get_at_pixel(x, y)

        # "self" placing turn
        if self.mode == 0:
            target_hex = self.hex_map.get_at_pixel(x, y)
            if target_hex:
                if self.reinforcement_troops > 0:
                    if (target_hex.properties["owner"] == "self"):
                        self.reinforcement_troops -= 1
                        target_hex.properties["troops"] += 1
                        self.troops_placed = True

                elif self.troops_placed:
                    self.mode = 1
                    self.troops_placed = False
                    self.reinforcement_troops = 0

        # "self" attacking turn
        elif self.mode == 1:
            if target_hex:
                if target_hex.properties["owner"] == "self":
                    self.attack_A = target_hex

                elif self.attack_A and (target_hex.properties["owner"] == "enemy"):
                    neighbors = self.get_neighbors(self.attack_A)
                    if target_hex in neighbors:
                        self.attack_B = target_hex
                        # go!
                        self.do_attack()

    def on_mouse_motion(self, x, y, dx, dy):
        target_hex = self.hex_map.get_at_pixel(x, y)

        # "self" attacking turn
        if self.mode == 1:
            if target_hex and self.attack_A:
                if target_hex:
                    neighbors = self.get_neighbors(self.attack_A)

                    if (target_hex.properties["owner"] == "enemy") and (target_hex in neighbors):
                        self.focused_cell = target_hex

                    else:
                        self.focused_cell = None

    def get_neighbors(self, cell):
        '''
            UP         -> (0, 1)
            DOWN       -> (0, -1)

            EVEN X
            -------
            UP RIGHT   -> (1, 0)
            UP LEFT    -> (-1, 0)
            DOWN RIGHT -> (1, -1)
            DOWN LEFT  -> (-1, -1)

            EVEN X
            -------
            UP RIGHT   -> (1, 1)
            UP LEFT    -> (-1, 1)
            DOWN RIGHT -> (1, 0)
            DOWN LEFT  -> (-1, 0)
        '''

        neighbors = []

        # UP
        try:
            neighbors.append(self.hex_map.cells[cell.i + 0][cell.j + 1])

        except IndexError:
            pass

        # DOWN
        try:
            neighbors.append(self.hex_map.cells[cell.i + 0][cell.j - 1])

        except IndexError:
            pass

        # UP RIGHT
        try:
            if cell.i % 2 == 0:
                neighbors.append(self.hex_map.cells[cell.i + 1][cell.j + 0])

            else:
                neighbors.append(self.hex_map.cells[cell.i + 1][cell.j + 1])

        except IndexError:
            pass

        # UP LEFT
        try:
            if cell.i % 2 == 0:
                neighbors.append(self.hex_map.cells[cell.i - 1][cell.j + 0])

            else:
                neighbors.append(self.hex_map.cells[cell.i - 1][cell.j + 1])

        except IndexError:
            pass

        # DOWN RIGHT
        try:
            if cell.i % 2 == 0:
                neighbors.append(self.hex_map.cells[cell.i + 1][cell.j - 1])

            else:
                neighbors.append(self.hex_map.cells[cell.i + 1][cell.j + 0])

        except IndexError:
            pass

        # DOWN LEFT
        try:
            if cell.i % 2 == 0:
                neighbors.append(self.hex_map.cells[cell.i - 1][cell.j - 1])

            else:
                neighbors.append(self.hex_map.cells[cell.i - 1][cell.j + 0])

        except IndexError:
            pass

        return neighbors

    def get_certain_neighbors(self, cell, owner):
        neighbors = self.get_neighbors(cell)

        certain_neighbors = []
        for tile in neighbors:
            if tile.properties["owner"] == owner:
                certain_neighbors.append(tile)

        return certain_neighbors

    def get_neighbor_positions(self, cell):
        '''
            UP         -> (0, 1)
            DOWN       -> (0, -1)
            UP RIGHT   -> (1, 1)
            UP LEFT    -> (-1, 1)
            DOWN RIGHT -> (1, 0)
            DOWN LEFT  -> (-1, 0)
        '''

        positions = [
            [cell.i + 0, cell.j + 1],
            [cell.i + 0, cell.j - 1],
            [cell.i + 1, cell.j + 1],
            [cell.i - 1, cell.j + 1],
            [cell.i + 1, cell.j + 0],
            [cell.i - 1, cell.j + 0]
        ]

        return positions

    def generate_territories(self):
        # (i, j) => (x, y)
        # first set all the tiles to "self" owned and 1 troop per cell
        for i in range(0, len(self.hex_map.cells)):
            for j in range(0, len(self.hex_map.cells[i])):
                self.hex_map.cells[i][j].properties["owner"] = "self"
                self.hex_map.cells[i][j].properties["troops"] = 1
                self.hex_map.cells[i][j].properties["focused"] = False

        # then generate the enemy's territory over that map
        last = int(self.difficulty_modifier * len(self.hex_map.cells))
        for j in range(0, len(self.hex_map.cells[0])):
            length = random.randint(last - self.random_range, last + self.random_range)
            for i in range(0, len(self.hex_map.cells)):
                if i > length:
                    break

                self.hex_map.cells[i][j].properties["owner"] = "enemy"

    # divy out the reinforcement troops at the beginning of each turn
    def get_reinforcement_troops(self, owner):
        reinforcement_troops = 0
        for i in range(0, len(self.hex_map.cells)):
            for j in range(0, len(self.hex_map.cells[i])):
                if self.hex_map.cells[i][j].properties["owner"] == owner:
                    reinforcement_troops += 1

        reinforcement_troops = int(reinforcement_troops * self.reinforcement_troops_modifier)
        return reinforcement_troops

    def do_attack(self):
        # do attack action
        self_troops = self.attack_A.properties["troops"]
        enemy_troops = self.attack_B.properties["troops"]
        attack = self_troops - enemy_troops
        if attack > 0:
            self.attack_A.properties["troops"] = 1
            self.attack_B.properties["owner"] = "self"
            self.attack_B.properties["troops"] = attack

        elif attack <= 0:
            self.attack_A.properties["troops"] = 1
            x = enemy_troops - self_troops
            if x == 0:
                x = 1

            self.attack_B.properties["troops"] = x

        # reset
        self.attack_A = None
        self.attack_B = None
        self.focused_cell = None

    '''def check_win(self):
        print("check_win")
        for i in range(0, len(self.hex_map)):
            for j in range(0, len(self.hex_map[i])):
                if self.hex_map.cells[i][j].properties["owner"] == "enemy":
                    print("found an enemy at ({}, {})".format(i, j))
                    return False

        print("You won.")
        return True

    def check_lose(self):
        print("check_lose")
        self_tiles = []
        for i in range(0, len(self.hex_map)):
            for j in range(0, len(self.hex_map[i])):
                if self.hex_map.cells[i][j].properties["owner"] == "self":
                    print("found a self at ({}, {})".format(i, j))
                    return False

        print("You lost.")
        return True'''

    def do_enemy_turn(self):
        self.do_enemy_reinforcements()
        for i in range(0, self.difficulty):
            self.do_enemy_attack()

    def do_enemy_reinforcements(self):
        reinforcements = self.get_reinforcement_troops("enemy")
        edge_tiles = self.get_edge_tiles()

        print("Enemy reinforcements this turn: {}".format(reinforcements))

        for tile in edge_tiles:
            if reinforcements < 1:
                break

            elif reinforcements < 3:
                tile.properties["troops"] += reinforcements
                reinforcements = 0
                break

            x = random.randint(1, int(reinforcements / 2))
            tile.properties["troops"] += x
            reinforcements -= x

    def do_enemy_attack(self):
        attacking_tiles = self.get_attacking_tiles()

        for tile in attacking_tiles:
            if random.randint(0, 1) == 0:
                neighbors = self.get_certain_neighbors(tile, "self")
                if len(neighbors) < 1:
                    continue

                i = random.randint(0, len(neighbors) - 1)

                self_troops = neighbors[i].properties["troops"]
                enemy_troops = tile.properties["troops"]
                attack = enemy_troops - self_troops
                if attack > 0:
                    tile.properties["troops"] = 1
                    neighbors[i].properties["owner"] = "enemy"
                    neighbors[i].properties["troops"] = attack

                elif attack <= 0:
                    tile.properties["troops"] = 1
                    x = self_troops - enemy_troops
                    if x == 0:
                        x = 1

                    neighbors[i].properties["troops"] = x

    def get_attacking_tiles(self):
        attacking_tiles = []
        for i in range(0, len(self.hex_map.cells)):
            for j in range(0, len(self.hex_map.cells[i])):
                if (self.hex_map.cells[i][j].properties["owner"] == "enemy") and (self.hex_map.cells[i][j].properties["troops"] > 1):
                    attacking_tiles.append(self.hex_map.cells[i][j])

        l = []
        edge_tiles = self.get_edge_tiles()
        for tile in attacking_tiles:
            if tile in edge_tiles:
                l.append(tile)

        return l

    def get_edge_tiles(self):
        edge_tiles = []
        for i in range(1, len(self.hex_map.cells) - 1):
            for j in range(1, len(self.hex_map.cells[i]) - 1):
                if self.hex_map.cells[i][j].properties["owner"] == "enemy":
                    if self.get_certain_neighbors(self.hex_map.cells[i][j], "self"):
                        edge_tiles.append(self.hex_map.cells[i][j])

        return edge_tiles

    def update(self, delta_time):
        # check if the player won or lost
        '''if self.check_lose:
            self.scene_manager.do_lose()

        elif self.check_win:
            self.scene_manager.do_win()'''

        # mode stuff
        if self.mode == 2:
            self.do_enemy_turn()
            # end enemy turn
            self.mode = 0
            self.reinforcement_troops = self.get_reinforcement_troops("self")
            '''if self.reinforcement_troops == 0:
                self.scene_manager.do_lose()'''

        # update the troop labels
        self.troop_layer.update_labels()

        # update the colors of each tile based on their "owner" property ("enemy" (dark) v.s. "self" (light))
        for i in range(0, len(self.hex_map.cells)):
            for j in range(0, len(self.hex_map.cells[i])):
                if self.hex_map.cells[i][j] == self.focused_cell:
                    self.hex_map.set_cell_color(i, j, self.focused_color)

                elif self.hex_map.cells[i][j] == self.attack_A:
                    self.hex_map.set_cell_color(i, j, (200, 200, 200))

                elif self.hex_map.cells[i][j].properties["owner"] == "enemy":
                    self.hex_map.set_cell_color(i, j, self.enemy_color)

                else:
                    self.hex_map.set_cell_color(i, j, (255, 255, 255))


class LevelScene(cocos.scene.Scene):
    def __init__(self, level_layer, troop_layer):
        super(LevelScene, self).__init__()

        self.level_layer = level_layer
        self.troop_layer = troop_layer

        self.level_layer.get_troop_layer(self.troop_layer)
        self.troop_layer.get_level_layer(self.level_layer)

        self.level_layer.start()
        self.troop_layer.start()

        self.add(self.level_layer)
        self.add(self.troop_layer)


'''class WinLayer(cocos.layer.Layer):
    def __init__(self):
        super(WinLayer, self).__init__()

        label = cocos.text.Label(
            "Congrats!  You won!",
            font_name = "Calibri",
            font_size = 64,
            anchor_x = "center", anchor_y = "center"
        )
        label.position = int(config.SCREEN_WIDTH / 2), int(config.SCREEN_HEIGHT / 2)

        label2 = cocos.text.Label(
            "Go back and raise the level difficulty!  The territory is randomly different each time!",
            font_name = "Calibri",
            font_size = 32,
            anchor_x = "center", anchor_y = "center"
        )
        label2.position = int(config.SCREEN_WIDTH / 2), int(config.SCREEN_HEIGHT / 2) - 50

        self.add(label)
        self.add(label2)


class LoseLayer(cocos.layer.Layer):
    def __init__(self):
        super(LoseLayer, self).__init__()

        label = cocos.text.Label(
            "Sorry, you lost.  Try again!",
            font_name = "Calibri",
            font_size = 64,
            anchor_x = "center", anchor_y = "center"
        )
        label.position = int(config.SCREEN_WIDTH / 2), int(config.SCREEN_HEIGHT)

        self.add(label)'''


def main():
    director.init(width = config.SCREEN_WIDTH, height = config.SCREEN_HEIGHT, resizable = False, caption = "Bonaparte!")

    # ONLY FOR DEV
    # director.show_FPS = True

    scene_manager = scenes.SceneManager(director)
    scene_manager.load_scenes(MainMenuScene(scene_manager, title = "Bonaparte!")) # , cocos.scene.Scene(WinLayer), cocos.scene.Scene(LoseLayer))
    scene_manager.load_levels(load_levels())
    scene_manager.run()

if __name__ == "__main__":
    main()
